__all__ = ["CombinedParser"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.parsers.parser import *


class CombinedParser(ParserInterface):
    def __init__(self, *parsers: Parser):
        self.parsers = parsers

    def parse(self, data_splitter=None, idmap: IDMap = None, show_pbar: bool = True):
        data_splitter = data_splitter or RandomSplitter([0.8, 0.2])
        idmap = idmap or IDMap()

        parsers_records = [
            o.parse_dicted(idmap=idmap, show_pbar=show_pbar) for o in self.parsers
        ]

        ids = [set(o.keys()) for o in parsers_records]
        valid_ids = set.intersection(*ids)
        excluded = set.union(*ids) - valid_ids
        print(f"Removed {len(excluded)} files: {excluded}")

        intersect_idmap = idmap.filter_ids(ids=valid_ids)
        splits = data_splitter(idmap=intersect_idmap)

        # TODO: Confusing names
        # combine keys for all parsers
        split_records = []
        for ids in splits:
            current_records = []
            for id in ids:
                record = {"imageid": id}
                for records in parsers_records:
                    record.update(records[id])
                current_records.append(record)
            split_records.append(current_records)

        return split_records



__all__ = ["convert_dataloader_to_fastai"]

from icevision.imports import *
from icevision.engines.fastai.imports import *
from torch.utils.data import SequentialSampler, RandomSampler


def convert_dataloader_to_fastai(dataloader: DataLoader):
    def raise_error_convert(data):
        raise NotImplementedError

    class FastaiDataLoaderWithCollate(fastai.DataLoader):
        def create_batch(self, b):
            return (dataloader.collate_fn, raise_error_convert)[self.prebatched](b)

    # use the type of sampler to determine if shuffle is true or false
    if isinstance(dataloader.sampler, SequentialSampler):
        shuffle = False
    elif isinstance(dataloader.sampler, RandomSampler):
        shuffle = True
    else:
        raise ValueError(
            f"Sampler {type(dataloader.sampler)} not supported. Fastai only"
            "supports RandomSampler or SequentialSampler"
        )

    return FastaiDataLoaderWithCollate(
        dataset=dataloader.dataset,
        bs=dataloader.batch_size,
        num_workers=dataloader.num_workers,
        drop_last=dataloader.drop_last,
        shuffle=shuffle,
        pin_memory=dataloader.pin_memory,
    )



__all__ = ["get_data_dir"]

from icevision.imports import *

root_dir = Path.home() / ".icevision"
root_dir.mkdir(exist_ok=True)

data_dir = root_dir / "data"
data_dir.mkdir(exist_ok=True)


def get_data_dir():
    return data_dir



__all__ = ["CaptureStdout"]

from icevision.imports import *
from io import StringIO


class CaptureStdout(list):
    """Capture the stdout (like prints)
    From: https://stackoverflow.com/a/16571630/6772672
    """

    def __init__(self, propagate_stdout: bool = False):
        self.propagate_stdout = propagate_stdout

    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout

        if self.propagate_stdout:
            print("\n".join(self))



__all__ = ["RCNNCallback"]

from icevision.imports import *
from icevision.engines.fastai import *
from icevision.models.rcnn import faster_rcnn


class RCNNCallback(fastai.Callback, ABC):
    @abstractmethod
    def convert_raw_predictions(self, raw_preds):
        """Convert raw predictions from the model to library standard."""

    def before_batch(self):
        assert len(self.xb) == len(self.yb) == 1, "Only works for single input-output"
        self.learn.xb = self.xb[0]
        self.learn.records = self.yb[0]
        self.learn.yb = ()

    def after_pred(self):
        self.learn.yb = [self.learn.xb[1]]
        self.learn.xb = [self.learn.xb[0]]

    def before_validate(self):
        # put model in training mode so we can calculate losses for validation
        self.model.train()

    def after_loss(self):
        if not self.training:
            self.model.eval()
            self.learn.pred = self.model(*self.xb)
            self.model.train()

            preds = self.convert_raw_predictions(raw_preds=self.pred)
            self.learn.converted_preds = preds



__all__ = ["open_img", "show_img", "plot_grid"]

from icevision.imports import *


def open_img(fn, gray=False):
    if not os.path.exists(fn):
        raise ValueError(f"File {fn} does not exists")
    color = cv2.COLOR_BGR2GRAY if gray else cv2.COLOR_BGR2RGB
    return cv2.cvtColor(cv2.imread(str(fn)), color)


def show_img(img, ax=None, show: bool = False, **kwargs):
    img = img.squeeze().copy()
    cmap = "gray" if len(img.shape) == 2 else None

    if ax is None:
        fig, ax = plt.subplots(**kwargs)

    ax.imshow(img, cmap=cmap)
    ax.set_axis_off()

    if show:
        plt.show()

    return ax


def plot_grid(fs: List[callable], ncols=1, figsize=None, show=False, **kwargs):
    figsize = figsize or (12 * len(fs) / ncols, 12)
    nrows = math.ceil(len(fs) / ncols)

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)

    axs = np.asarray(axs)
    for f, ax in zip(fs, axs.flatten()):
        f(ax=ax)

    plt.tight_layout()
    if show:
        plt.show()



__all__ = ["predict", "predict_dl", "convert_raw_prediction", "convert_raw_predictions"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.models.utils import _predict_dl
from icevision.models.rcnn.faster_rcnn.prediction import (
    convert_raw_prediction as faster_convert_raw_prediction,
)


@torch.no_grad()
def predict(
    model: nn.Module,
    batch: List[torch.Tensor],
    detection_threshold: float = 0.5,
    mask_threshold: float = 0.5,
    device: Optional[torch.device] = None,
):
    model.eval()
    device = device or model_device(model)
    batch = [o.to(device) for o in batch]

    raw_preds = model(*batch)
    return convert_raw_predictions(
        raw_preds=raw_preds,
        detection_threshold=detection_threshold,
        mask_threshold=mask_threshold,
    )


def predict_dl(
    model: nn.Module, infer_dl: DataLoader, show_pbar: bool = True, **predict_kwargs
):
    return _predict_dl(
        predict_fn=predict,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(
    raw_preds: List[dict], detection_threshold: float, mask_threshold: float
):
    return [
        convert_raw_prediction(
            raw_pred=raw_pred,
            detection_threshold=detection_threshold,
            mask_threshold=mask_threshold,
        )
        for raw_pred in raw_preds
    ]


def convert_raw_prediction(
    raw_pred: dict, detection_threshold: float, mask_threshold: float
):
    preds = faster_convert_raw_prediction(
        raw_pred=raw_pred, detection_threshold=detection_threshold
    )

    above_threshold = preds["above_threshold"]
    masks_probs = raw_pred["masks"][above_threshold]
    masks_probs = masks_probs.detach().cpu().numpy()
    # convert probabilities to 0 or 1 based on mask_threshold
    masks = masks_probs > mask_threshold
    masks = MaskArray(masks.squeeze(1))

    return {"masks": masks, **preds}



__all__ = ["voc", "VocXmlParser", "VocMaskParser"]

import xml.etree.ElementTree as ET
from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.parsers.parser import *
from icevision.parsers.defaults import *
from icevision.parsers.mixins import *


def voc(
    annotations_dir: Union[str, Path],
    images_dir: Union[str, Path],
    class_map: ClassMap,
    mask: bool = False,
):
    parser = VocXmlParser(
        annotations_dir=annotations_dir,
        images_dir=images_dir,
        class_map=class_map,
    )

    if mask:
        mask_parser = VocMaskParser(data_dir / "annotations/trimaps")
        parser = CombinedParser(parser, mask_parser)

    return parser


class VocXmlParser(DefaultImageInfoParser, LabelsMixin, BBoxesMixin):
    def __init__(
        self,
        annotations_dir: Union[str, Path],
        images_dir: Union[str, Path],
        class_map: ClassMap,
    ):
        self.images_dir = Path(images_dir)
        self.class_map = class_map

        self.annotations_dir = Path(annotations_dir)
        self.annotation_files = get_files(self.annotations_dir, extensions=[".xml"])

    def __len__(self):
        return len(self.annotation_files)

    def __iter__(self):
        yield from self.annotation_files

    def prepare(self, o):
        tree = ET.parse(str(o))
        self._root = tree.getroot()
        self._filename = self._root.find("filename").text
        self._size = self._root.find("size")

    def imageid(self, o) -> Hashable:
        return str(Path(self._filename).stem)

    def filepath(self, o) -> Union[str, Path]:
        return self.images_dir / self._filename

    def image_height(self, o) -> int:
        return int(self._size.find("height").text)

    def image_width(self, o) -> int:
        return int(self._size.find("width").text)

    def labels(self, o) -> List[int]:
        labels = []
        for object in self._root.iter("object"):
            label = object.find("name").text
            label_id = self.class_map.get_name(label)
            labels.append(label_id)

        return labels

    def bboxes(self, o) -> List[BBox]:
        def to_int(x):
            return int(float(x))

        bboxes = []
        for object in self._root.iter("object"):
            xml_bbox = object.find("bndbox")
            xmin = to_int(xml_bbox.find("xmin").text)
            ymin = to_int(xml_bbox.find("ymin").text)
            xmax = to_int(xml_bbox.find("xmax").text)
            ymax = to_int(xml_bbox.find("ymax").text)

            bbox = BBox.from_xyxy(xmin, ymin, xmax, ymax)
            bboxes.append(bbox)

        return bboxes


class VocMaskParser(Parser, ImageidMixin, MasksMixin):
    def __init__(self, masks_dir: Union[str, Path]):
        self.mask_files = get_image_files(masks_dir)

    def __len__(self):
        return len(self.mask_files)

    def __iter__(self):
        yield from self.mask_files

    def imageid(self, o) -> Hashable:
        return str(Path(o).stem)

    def masks(self, o) -> List[Mask]:
        return [VocMaskFile(o)]



__all__ = [
    "notnone",
    "ifnotnone",
    "last",
    "lmap",
    "allequal",
    "cleandict",
    "mergeds",
    "zipsafe",
    "np_local_seed",
    "pbar",
    "IMAGENET_STATS",
    "normalize",
    "denormalize",
    "normalize_imagenet",
    "denormalize_imagenet",
]

from icevision.imports import *


def notnone(x):
    return x is not None


def ifnotnone(x, f):
    return f(x) if notnone(x) else x


def last(x):
    return next(reversed(x))


def lmap(f, xs):
    return list(map(f, xs)) if notnone(xs) else None


def allequal(l):
    return l.count(l[0]) == len(l) if l else True


def cleandict(d):
    return {k: v for k, v in d.items() if notnone(v)}


def mergeds(ds):
    aux = defaultdict(list)
    for d in ds:
        for k, v in d.items():
            aux[k].append(v)
    return dict(aux)


def zipsafe(*its):
    if not allequal(lmap(len, its)):
        raise ValueError("The elements have different leghts")
    return zip(*its)


def pbar(iter, show=True):
    return tqdm(iter) if show else iter


@contextmanager
def np_local_seed(seed):
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)


def normalize(img, mean, std, max_pixel_value=255):
    img = img.astype(np.float32)
    img /= max_pixel_value

    mean, std = map(np.float32, [mean, std])
    return (img - mean) / std


def denormalize(img, mean, std, max_pixel_value=255):
    return np.around((img * std + mean) * max_pixel_value).astype(np.uint8)


IMAGENET_STATS = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])


def denormalize_imagenet(img):
    mean, std = IMAGENET_STATS
    return denormalize(img=img, mean=mean, std=std)


def normalize_imagenet(img):
    mean, std = IMAGENET_STATS
    return normalize(img=img, mean=mean, std=std)



from icevision.models.utils import *

from icevision.models.rcnn import faster_rcnn, mask_rcnn

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.effdet:
    from icevision.models import efficientdet



__all__ = [
    "create_coco_api",
    "convert_records_to_coco_style",
    "convert_preds_to_coco_style",
    "convert_record_to_coco_annotations",
    "coco_api_from_records",
    "coco_api_from_preds",
    "create_coco_eval",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval


def create_coco_api(coco_records) -> COCO:
    """Create COCO dataset api

    Args:
        coco_records: Records in coco style (use convert_records_to_coco_style to convert
        records to coco style.
    """
    coco_ds = COCO()
    coco_ds.dataset = coco_records
    coco_ds.createIndex()

    return coco_ds


def coco_api_from_preds(preds, show_pbar: bool = False) -> COCO:
    coco_preds = convert_preds_to_coco_style(preds, show_pbar=show_pbar)
    return create_coco_api(coco_preds)


def coco_api_from_records(records, show_pbar: bool = False) -> COCO:
    """Create pycocotools COCO dataset from records"""
    coco_records = convert_records_to_coco_style(records, show_pbar=show_pbar)
    return create_coco_api(coco_records=coco_records)


def create_coco_eval(
    records, preds, metric_type: str, show_pbar: bool = False
) -> COCOeval:
    assert len(records) == len(preds)

    for record, pred in zip(records, preds):
        pred["imageid"] = record["imageid"]
        pred["height"] = record["height"]
        pred["width"] = record["width"]
        # needs 'filepath' for mask `coco.py#418`
        pred["filepath"] = record["filepath"]

    target_ds = coco_api_from_records(records, show_pbar=show_pbar)
    pred_ds = coco_api_from_preds(preds, show_pbar=show_pbar)
    return COCOeval(target_ds, pred_ds, metric_type)


def convert_record_to_coco_image(record) -> dict:
    image = {}
    image["id"] = record["imageid"]
    image["file_name"] = Path(record["filepath"]).name
    image["width"] = record["width"]
    image["height"] = record["height"]
    return image


def convert_record_to_coco_annotations(record):
    annotations_dict = {
        "image_id": [],
        "category_id": [],
        "bbox": [],
        "area": [],
        "iscrowd": [],
    }
    # build annotations field
    for label in record["labels"]:
        annotations_dict["image_id"].append(record["imageid"])
        annotations_dict["category_id"].append(label)

    for bbox in record["bboxes"]:
        annotations_dict["bbox"].append(bbox.xywh)

    if "areas" in record:
        for area in record["areas"]:
            annotations_dict["area"].append(area)
    else:
        for bbox in record["bboxes"]:
            annotations_dict["area"].append(bbox.area)

    # HACK: Because of prepare_record, mask should always be `MaskArray`,
    # maybe the for loop is not required?
    if "masks" in record:
        annotations_dict["segmentation"] = []
        for mask in record["masks"]:
            if isinstance(mask, MaskArray):
                # HACK: see previous hack
                assert len(mask.shape) == 2
                mask2 = MaskArray(mask.data[None])
                rles = mask2.to_coco_rle(record["height"], record["width"])
                annotations_dict["segmentation"].extend(rles)
            elif isinstance(mask, Polygon):
                annotations_dict["segmentation"].append(mask.points)
            elif isinstance(mask, RLE):
                coco_rle = {
                    "counts": mask.to_coco(),
                    "size": [record["height"], record["width"]],
                }
                annotations_dict["segmentation"].append(coco_rle)
            elif isinstance(mask, MaskFile):
                rles = mask.to_coco_rle(record["height"], record["width"])
                annotations_dict["segmentation"].extend(rles)
            else:
                msg = f"Mask type {type(mask)} unsupported, we only support RLE and Polygon"
                raise ValueError(msg)

    # TODO: is auto assigning a value for iscrowds dangerous (may hurt the metric value?)
    if "iscrowds" not in record:
        record["iscrowds"] = [0] * len(record["labels"])
    for iscrowd in record["iscrowds"]:
        annotations_dict["iscrowd"].append(iscrowd)

    if "scores" in record:
        annotations_dict["score"] = record["scores"]

    return annotations_dict


def convert_preds_to_coco_style(preds, show_pbar: bool = False):
    return convert_records_to_coco_style(
        records=preds, images=True, categories=False, show_pbar=show_pbar
    )


def convert_records_to_coco_style(
    records,
    images: bool = True,
    annotations: bool = True,
    categories: bool = True,
    show_pbar: bool = True,
):
    """Converts records from library format to coco format.
    Inspired from: https://github.com/pytorch/vision/blob/master/references/detection/coco_utils.py#L146
    """
    images_ = []
    annotations_dict = defaultdict(list)

    for record in pbar(records, show=show_pbar):
        if images:
            image_ = convert_record_to_coco_image(record)
            images_.append(image_)

        if annotations:
            annotations_ = convert_record_to_coco_annotations(record)
            for k, v in annotations_.items():
                annotations_dict[k].extend(v)

    if annotations:
        annotations_dict = dict(annotations_dict)
        if not allequal([len(o) for o in annotations_dict.values()]):
            raise RuntimeError("Mismatch lenght of elements")

        # convert dict of lists to list of dicts
        annotations_ = []
        for i in range(len(annotations_dict["image_id"])):
            annotation = {k: v[i] for k, v in annotations_dict.items()}
            # annotations should be initialized starting at 1 (torchvision issue #1530)
            annotation["id"] = i + 1
            annotations_.append(annotation)

    if categories:
        categories_set = set(annotations_dict["category_id"])
        categories_ = [{"id": i} for i in categories_set]

    res = {}
    if images_:
        res["images"] = images_
    if annotations:
        res["annotations"] = annotations_
    if categories:
        res["categories"] = categories_

    return res



__all__ = ["LightningModelAdapter"]

import pytorch_lightning as pl
from icevision.imports import *
from icevision.metrics import *


class LightningModelAdapter(pl.LightningModule, ABC):
    def __init__(self, metrics: List[Metric] = None):
        super().__init__()
        self.metrics = metrics or []

    def accumulate_metrics(self, records, preds):
        for metric in self.metrics:
            metric.accumulate(records=records, preds=preds)

    def finalize_metrics(self) -> dict:
        all_logs = {}
        for metric in self.metrics:
            metric_logs = metric.finalize()
            metric_logs = {f"{metric.name}/{k}": v for k, v in metric_logs.items()}
            all_logs.update(metric_logs)
        return all_logs



__all__ = ["base_show_results"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.visualize import *
from icevision.data import *


def base_show_results(
    predict_fn: callable,
    build_infer_batch_fn: callable,
    model: nn.Module,
    dataset: Dataset,
    class_map: Optional[ClassMap] = None,
    num_samples: int = 6,
    ncols: int = 3,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    show: bool = True,
) -> None:
    samples = [dataset[i] for i in range(num_samples)]
    batch, samples = build_infer_batch_fn(samples)
    preds = predict_fn(model, batch)

    imgs = [sample["img"] for sample in samples]
    show_preds(
        imgs,
        preds,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        ncols=ncols,
        show=show,
    )



__all__ = ["EfficientDetCallback"]

from icevision.models import efficientdet
from icevision.engines.fastai import *


class EfficientDetCallback(fastai.Callback):
    def before_batch(self):
        assert len(self.xb) == len(self.yb) == 1, "Only works for single input-output"
        self.learn.xb = self.xb[0]
        self.learn.records = self.yb[0]
        self.learn.yb = ()

    def after_pred(self):
        self.learn.yb = [self.learn.xb[1]]
        self.learn.xb = [self.learn.xb[0]]

        if not self.training:
            preds = efficientdet.convert_raw_predictions(self.pred["detections"], 0)
            self.learn.converted_preds = preds



__all__ = [
    "annotations_dict",
    "sample_image_info_parser",
    "sample_annotation_parser",
    "sample_combined_parser",
    "sample_records",
    "sample_dataset",
]

import icevision
from icevision.imports import *
from icevision import *

try:
    source = Path(icevision.__file__).parent.parent / "samples"
    annotations_dict = json.loads((source / "annotations.json").read())
except FileNotFoundError:
    annotations_dict = None


def sample_image_info_parser():
    return parsers.COCOImageInfoParser(
        infos=annotations_dict["images"], img_dir=source / "images"
    )


def sample_annotation_parser():
    return parsers.COCOAnnotationParser(annotations=annotations_dict["annotations"])


def sample_combined_parser():
    return parsers.CombinedParser(
        sample_image_info_parser(), sample_annotation_parser()
    )


def sample_records():
    parser = sample_combined_parser()
    data_splitter = RandomSplitter([0.8, 0.2], seed=42)
    return parser.parse(data_splitter)


def sample_dataset():
    train_rs, valid_rs = sample_records()
    return Dataset(train_rs)



__all__ = ["Adapter", "aug_tfms", "resize_and_pad"]

import albumentations as A
from icevision.imports import *
from icevision.core import *
from icevision.tfms.transform import *


def _resize(size, ratio_resize=A.LongestMaxSize):
    return ratio_resize(size) if isinstance(size, int) else A.Resize(*size)


def resize_and_pad(
    size: Union[int, Tuple[int, int]],
    pad: A.DualTransform = partial(
        A.PadIfNeeded, border_mode=cv2.BORDER_CONSTANT, value=[124, 116, 104]
    ),
):
    height, width = (size, size) if isinstance(size, int) else size
    return [_resize(size), pad(min_height=height, min_width=width)]


def aug_tfms(
    size: Union[int, Tuple[int, int]],
    presize: Optional[Union[int, Tuple[int, int]]] = None,
    horizontal_flip: Optional[A.HorizontalFlip] = A.HorizontalFlip(),
    shift_scale_rotate: Optional[A.ShiftScaleRotate] = A.ShiftScaleRotate(),
    rgb_shift: Optional[A.RGBShift] = A.RGBShift(),
    lightning: Optional[A.RandomBrightnessContrast] = A.RandomBrightnessContrast(),
    blur: Optional[A.Blur] = A.Blur(blur_limit=(1, 3)),
    crop_fn: Optional[A.DualTransform] = partial(A.RandomSizedBBoxSafeCrop, p=0.5),
    pad: Optional[A.DualTransform] = partial(
        A.PadIfNeeded, border_mode=cv2.BORDER_CONSTANT, value=[124, 116, 104]
    ),
) -> List[A.BasicTransform]:
    """Collection of useful augmentation transforms.

    # Arguments
        size: The final size of the image. If an `int` is given, the maximum size of
            the image is rescaled, maintaing aspect ratio. If a `tuple` is given,
            the image is rescaled to have that exact size (height, width).
        presizing: Rescale the image before applying other transfroms. If `None` this
                transform is not applied. First introduced by fastai,this technique is
                explained in their book in [this](https://github.com/fastai/fastbook/blob/master/05_pet_breeds.ipynb)
                chapter (tip: search for "Presizing").
        horizontal_flip: Flip around the y-axis. If `None` this transform is not applied.
        shift_scale_rotate: Randomly shift, scale, and rotate. If `None` this transform
                is not applied.
        rgb_shift: Randomly shift values for each channel of RGB image. If `None` this
                transform is not applied.
        lightning: Randomly changes Brightness and Contrast. If `None` this transform
                is not applied.
        blur: Randomly blur the image. If `None` this transform is not applied.
        crop_fn: Randomly crop the image. If `None` this transform is not applied.
                Use `partial` to saturate other parameters of the class.
        pad: Pad the image to `size`, squaring the image if `size` is an `int`.
            If `None` this transform is not applied. Use `partial` to sature other
            parameters of the class.

    # Returns
        A list of albumentations transforms.
    """

    height, width = (size, size) if isinstance(size, int) else size

    tfms = []
    tfms += [_resize(presize, A.SmallestMaxSize) if presize is not None else None]
    tfms += [horizontal_flip, shift_scale_rotate, rgb_shift, lightning, blur]
    # Resize as the last transforms to reduce the number of artificial artifacts created
    if crop_fn is not None:
        crop = crop_fn(height=height, width=width)
        tfms += [A.OneOrOther(crop, _resize(size), p=crop.p)]
    else:
        tfms += [_resize(size)]
    tfms += [pad(min_height=height, min_width=width) if pad is not None else None]

    tfms = [tfm for tfm in tfms if tfm is not None]

    return tfms


class Adapter(Transform):
    """Adapter that enables the use of albumentations transforms.

    # Arguments
        tfms: `Sequence` of albumentation transforms.
    """

    def __init__(self, tfms: Sequence[A.BasicTransform]):
        self.bbox_params = A.BboxParams(format="pascal_voc", label_fields=["labels"])
        super().__init__(tfms=A.Compose(tfms, bbox_params=self.bbox_params))

    def apply(
        self,
        img: np.ndarray,
        labels=None,
        bboxes: List[BBox] = None,
        masks: MaskArray = None,
        iscrowds: List[int] = None,
        **kwargs
    ):
        # Substitue labels with list of idxs, so we can also filter out iscrowds in case any bboxes is removed
        # TODO: Same should be done if a masks is completely removed from the image (if bboxes is not given)
        params = {"image": img}
        params["labels"] = list(range_of(labels)) if labels is not None else []
        params["bboxes"] = [o.xyxy for o in bboxes] if bboxes is not None else []
        if masks is not None:
            params["masks"] = masks.data

        d = self.tfms(**params)

        out = {"img": d["image"]}
        out["height"], out["width"], _ = out["img"].shape

        # We use the values in d['labels'] to get what was removed by the transform
        if labels is not None:
            out["labels"] = [labels[i] for i in d["labels"]]
        if bboxes is not None:
            out["bboxes"] = [BBox.from_xyxy(*points) for points in d["bboxes"]]
        if masks is not None:
            keep_masks = [d["masks"][i] for i in d["labels"]]
            out["masks"] = MaskArray(np.array(keep_masks))
        if iscrowds is not None:
            out["iscrowds"] = [iscrowds[i] for i in d["labels"]]
        return out



__all__ = ["FastaiMetricAdapter"]

from icevision.imports import *
from icevision.metrics import Metric
from icevision.engines.fastai.imports import *


class FastaiMetricAdapter(fastai.Metric):
    def __init__(self, metric: Metric):
        self.metric = metric

    def reset(self):
        pass

    def accumulate(self, learn: fastai.Learner):
        self.metric.accumulate(records=learn.records, preds=learn.converted_preds)

    @property
    def value(self) -> Dict[str, float]:
        # return self.metric.finalize()
        # HACK: Return single item from dict
        logs = self.metric.finalize()
        return next(iter(logs.values()))

    @property
    def name(self) -> str:
        return self.metric.name



__all__ = [
    "show_sample",
    "show_record",
    "show_pred",
    "show_samples",
    "show_records",
    "show_preds",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.parsers import *
from icevision.data import *
from icevision.visualize.draw_data import *


def show_sample(
    sample,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    show: bool = True,
    ax: plt.Axes = None,
) -> None:
    img = draw_sample(
        sample=sample,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
    )
    show_img(img=img, ax=ax, show=show)


def show_record(
    record,
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    ax: plt.Axes = None,
    show: bool = False,
    prepare_record: Optional[callable] = None,
) -> None:
    img = draw_record(
        record=record,
        class_map=class_map,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        prepare_record=prepare_record,
    )
    show_img(img=img, ax=ax, show=show)


def show_pred(
    img: np.ndarray,
    pred: dict,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    show: bool = True,
    ax: plt.Axes = None,
) -> None:
    img = draw_pred(
        img=img,
        pred=pred,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
    )
    show_img(img=img, ax=ax, show=show)


def show_records(
    records: Sequence[RecordType],
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    prepare_record: Optional[callable] = None,
    ncols: int = 1,
    figsize=None,
    show: bool = False,
) -> None:
    partials = [
        partial(
            show_record,
            record=record,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            class_map=class_map,
            prepare_record=prepare_record,
            show=False,
        )
        for record in records
    ]
    plot_grid(partials, ncols=ncols, figsize=figsize, show=show)


def show_samples(
    samples: Sequence[dict],
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    ncols: int = 1,
    figsize=None,
    show=False,
) -> None:
    partials = [
        partial(
            show_sample,
            sample=sample,
            denormalize_fn=denormalize_fn,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            class_map=class_map,
            show=False,
        )
        for sample in samples
    ]
    plot_grid(partials, ncols=ncols, figsize=figsize, show=show)


def show_preds(
    imgs: Sequence[np.ndarray],
    preds: Sequence[dict],
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    ncols: int = 1,
    figsize=None,
    show=False,
) -> None:
    if not len(imgs) == len(preds):
        raise ValueError(
            f"Number of imgs ({len(imgs)}) should be the same as "
            f"the number of preds ({len(preds)})"
        )

    partials = [
        partial(
            show_pred,
            img=img,
            pred=pred,
            class_map=class_map,
            denormalize_fn=denormalize_fn,
            display_label=display_label,
            display_bbox=display_bbox,
            display_mask=display_mask,
            show=False,
        )
        for img, pred in zip(imgs, preds)
    ]
    plot_grid(partials, ncols=ncols, figsize=figsize, show=show)



import cv2, json, mimetypes, torch, torchvision, dataclasses, zipfile

from dataclasses import dataclass
from fastcore.imports import *
from fastcore.foundation import *
from fastcore.utils import *

import numpy as np
import torch.nn as nn
import torch.optim.lr_scheduler as lr_scheduler
import matplotlib.pyplot as plt

from abc import (
    ABC,
    abstractmethod,
    abstractproperty,
    abstractclassmethod,
    abstractstaticmethod,
)
from collections import defaultdict, OrderedDict
from enum import Enum
from tqdm.auto import tqdm
from pycocotools import mask as mask_utils
from contextlib import contextmanager
from typing import *

from torch import tensor, Tensor
from torch.utils.data import DataLoader
from torch.optim import SGD, Adam, AdamW, Adagrad, Adadelta, RMSprop
from torch.optim.lr_scheduler import (
    LambdaLR,
    StepLR,
    MultiStepLR,
    MultiplicativeLR,
    OneCycleLR,
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
)

from torchvision.transforms.functional import to_tensor as im2tensor

from loguru import logger

# Soft imports
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import fastai.vision.all as fastai

if SoftDependencies.pytorch_lightning:
    import pytorch_lightning as pl


# TODO: Stop importing partial from fastcore and move this to utils
class partial:
    """Wraps functools.partial, same functionality.

    Modifies the original partial `__repr__` and `__str__` in other to fix #270
    """

    def __init__(self, func, *args, **kwargs):
        self._partial = functools.partial(func, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self._partial(*args, **kwargs)

    def __str__(self):
        name = self._partial.func.__name__
        partial_str = str(self._partial)
        return re.sub(r"<.+>", name, partial_str)

    def __repr__(self):
        return str(self)



__all__ = ["ModelAdapter"]

from icevision.imports import *
from icevision.utils import *
from icevision.metrics import *
from icevision.engines.lightning.lightning_model_adapter import LightningModelAdapter
from icevision.models import efficientdet


class ModelAdapter(LightningModelAdapter, ABC):
    """Lightning module specialized for EfficientDet, with metrics support.

    The methods `forward`, `training_step`, `validation_step`, `validation_epoch_end`
    are already overriden.

    # Arguments
        model: The pytorch model to use.
        metrics: `Sequence` of metrics to use.

    # Returns
        A `LightningModule`.
    """

    def __init__(self, model: nn.Module, metrics: List[Metric] = None):
        super().__init__(metrics=metrics)
        self.model = model

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        (xb, yb), records = batch
        preds = self(xb, yb)

        loss = efficientdet.loss_fn(preds, yb)
        log = {f"train/{k}": v for k, v in preds.items()}

        return {"loss": loss, "log": log}

    def validation_step(self, batch, batch_idx):
        (xb, yb), records = batch

        with torch.no_grad():
            raw_preds = self(xb, yb)
            preds = efficientdet.convert_raw_predictions(raw_preds["detections"], 0)
            loss = efficientdet.loss_fn(raw_preds, yb)

        self.accumulate_metrics(records, preds)

        log = {f"valid/{k}": v for k, v in raw_preds.items() if "loss" in k}
        return {"valid/loss": loss, "log": log}

    def validation_epoch_end(self, outs):
        merged_outs = mergeds(outs)
        avg_loss = torch.stack(merged_outs["valid/loss"]).mean()

        merged_logs = mergeds([out["log"] for out in outs])
        avg_logs = {k: torch.stack(v).mean() for k, v in merged_logs.items()}

        metrics_log = self.finalize_metrics()

        log = {**avg_logs, **metrics_log}
        return {"val_loss": avg_loss, "log": log}



__all__ = [
    "filter_params",
    "unfreeze",
    "freeze",
    "transform_dl",
    "common_build_batch",
    "_predict_dl",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.parsers import *


BN_TYPES = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)


def filter_params(
    module: nn.Module, bn: bool = True, only_trainable=False
) -> Generator:
    """Yields the trainable parameters of a given module.

    Args:
        module: A given module
        bn: If False, don't return batch norm layers

    Returns:
        Generator
    """
    children = list(module.children())
    if not children:
        if not isinstance(module, BN_TYPES) or bn:
            for param in module.parameters():
                if not only_trainable or param.requires_grad:
                    yield param
    else:
        for child in children:
            for param in filter_params(
                module=child, bn=bn, only_trainable=only_trainable
            ):
                yield param


def unfreeze(params):
    for p in params:
        p.requires_grad = True


def freeze(params):
    for p in params:
        p.requires_grad = False


def transform_dl(dataset, build_batch, batch_tfms=None, **dataloader_kwargs):
    collate_fn = partial(build_batch, batch_tfms=batch_tfms)
    return DataLoader(dataset=dataset, collate_fn=collate_fn, **dataloader_kwargs)


def common_build_batch(records: Sequence[RecordType], batch_tfms=None):
    if batch_tfms is not None:
        records = batch_tfms(records)

    return records


@torch.no_grad()
def _predict_dl(
    predict_fn,
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    **predict_kwargs,
):
    all_preds, all_samples = [], []
    for batch, samples in pbar(infer_dl, show=show_pbar):
        preds = predict_fn(model=model, batch=batch, **predict_kwargs)

        all_samples.extend(samples)
        all_preds.extend(preds)

    return all_samples, all_preds



__all__ = ["ImgPadStack"]

from icevision.imports import *
from icevision.core import *
from icevision.tfms.batch.batch_transform import BatchTransform


class ImgPadStack(BatchTransform):
    def __init__(self, pad_value: Union[float, Sequence[float]] = 0.0):
        # reshape makes sure array always have one dimension (for the single float case)
        self.pad_value = np.array(pad_value).reshape(-1)

    def apply(self, records: List[RecordType]) -> List[RecordType]:
        max_sizes = np.zeros(3, dtype=int)
        for record in records:
            max_sizes = np.maximum(max_sizes, record["img"].shape)

        img_dtype = records[0]["img"].dtype
        padded_imgs = np.ones((len(records), *max_sizes), dtype=img_dtype)
        padded_imgs *= self.pad_value
        for record, padded_img in zip(records, padded_imgs):
            img = record["img"]

            xmax, ymax, zmax = img.shape
            padded_img[:xmax, :ymax, :zmax] = img
            record["img"] = padded_img

        return records



__all__ = ["Transform"]

from icevision.imports import *
from icevision.core import *


class Transform(ABC):
    def __init__(self, tfms):
        self.tfms = tfms

    def __call__(self, data: dict):
        data = data.copy()
        tfmed = self.apply(**data)
        data.update(tfmed)
        return data

    @abstractmethod
    def apply(
        self,
        img: np.ndarray,
        imageid: int,
        label: List[int],
        iscrowd: List[int],
        bbox: List[BBox],
        mask: MaskArray,
    ):
        """Apply the transform
        Returns:
              dict: Modified values, the keys of the dictionary should have the same
              names as the keys received by this function
        """



__all__ = ["predict", "predict_dl", "convert_raw_prediction", "convert_raw_predictions"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.models.utils import _predict_dl


@torch.no_grad()
def predict(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
):
    model.eval()
    device = device or model_device(model)
    batch = [o.to(device) for o in batch]

    raw_preds = model(*batch)
    return convert_raw_predictions(
        raw_preds=raw_preds, detection_threshold=detection_threshold
    )


def predict_dl(
    model: nn.Module,
    infer_dl: DataLoader,
    show_pbar: bool = True,
    **predict_kwargs,
):
    return _predict_dl(
        predict_fn=predict,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(raw_preds, detection_threshold: float):
    return [
        convert_raw_prediction(raw_pred, detection_threshold=detection_threshold)
        for raw_pred in raw_preds
    ]


def convert_raw_prediction(raw_pred: dict, detection_threshold: float):
    above_threshold = raw_pred["scores"] >= detection_threshold

    labels = raw_pred["labels"][above_threshold]
    labels = labels.detach().cpu().numpy()

    scores = raw_pred["scores"][above_threshold]
    scores = scores.detach().cpu().numpy()

    boxes = raw_pred["boxes"][above_threshold]
    bboxes = []
    for box_tensor in boxes:
        xyxy = box_tensor.cpu().numpy()
        bbox = BBox.from_xyxy(*xyxy)
        bboxes.append(bbox)

    return {
        "labels": labels,
        "scores": scores,
        "bboxes": bboxes,
        "above_threshold": above_threshold,
    }



__all__ = [
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
    "train_dl",
    "valid_dl",
    "infer_dl",
]

from icevision.imports import *
from icevision.models.utils import *


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for training the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for validating the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def infer_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for inferring the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_infer_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def build_train_batch(records, batch_tfms=None):
    """Builds a batch in the format required by the model when training.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_train_batch(records)
    outs = model(*batch)
    ```
    """
    records = common_build_batch(records, batch_tfms=batch_tfms)

    images = []
    targets = {"bbox": [], "cls": []}
    for record in records:
        image = im2tensor(record["img"])
        images.append(image)

        if len(record["labels"]) == 0:
            targets["cls"].append(tensor([0], dtype=torch.float))
            targets["bbox"].append(tensor([[0, 0, 0, 0]], dtype=torch.float))
        else:
            labels = tensor(record["labels"], dtype=torch.float)
            targets["cls"].append(labels)

            bboxes = tensor([bbox.yxyx for bbox in record["bboxes"]], dtype=torch.float)
            targets["bbox"].append(bboxes)

    images = torch.stack(images)

    return (images, targets), records


def build_valid_batch(records, batch_tfms=None):
    """Builds a batch in the format required by the model when validating.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_valid_batch(records)
    outs = model(*batch)
    ```
    """
    (images, targets), records = build_train_batch(
        records=records, batch_tfms=batch_tfms
    )

    img_sizes = [(r["height"], r["width"]) for r in records]
    targets["img_size"] = tensor(img_sizes, dtype=torch.float)

    targets["img_scale"] = tensor([1] * len(records), dtype=torch.float)

    return (images, targets), records


def build_infer_batch(dataset, batch_tfms=None):
    """Builds a batch in the format required by the model when doing inference.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied. # Examples
    Use the result of this function to feed the model.
    ```python
    batch, records = build_infer_batch(records)
    outs = model(*batch)
    ```
    """
    samples = common_build_batch(dataset, batch_tfms=batch_tfms)

    tensor_imgs, img_sizes = [], []
    for record in samples:
        tensor_imgs.append(im2tensor(record["img"]))
        img_sizes.append((record["height"], record["width"]))

    tensor_imgs = torch.stack(tensor_imgs)
    tensor_sizes = tensor(img_sizes, dtype=torch.float)
    tensor_scales = tensor([1] * len(samples), dtype=torch.float)

    return (tensor_imgs, tensor_scales, tensor_sizes), samples



__all__ = ["default_prepare_record"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *


def prepare_copy(record):
    return record.copy()


def prepare_img(record):
    if "filepath" in record:
        record["img"] = open_img(record["filepath"])
    return record


def prepare_img_size(record):
    if "img" in record:
        record["height"], record["width"], _ = record["img"].shape
    return record


def prepare_mask(record):
    if "masks" in record:
        record["masks"] = MaskArray.from_masks(
            record["masks"], record["height"], record["width"]
        )
    return record


default_prepare_record = compose(
    prepare_copy,
    prepare_img,
    prepare_img_size,
    prepare_mask,
)



__all__ = ["model"]

from icevision.imports import *
from icevision.utils import *
from effdet import get_efficientdet_config, EfficientDet, DetBenchTrain, unwrap_bench
from effdet.efficientdet import HeadNet


def model(
    model_name: str, num_classes: int, img_size: int, pretrained: bool = True
) -> nn.Module:
    """Creates the efficientdet model specified by `model_name`.

    The model implementation is by Ross Wightman, original repo
    [here](https://github.com/rwightman/efficientdet-pytorch).

    # Arguments
        model_name: Specifies the model to create. For pretrained models, check
            [this](https://github.com/rwightman/efficientdet-pytorch#models) table.
        num_classes: Number of classes of your dataset (including background).
        img_size: Image size that will be fed to the model. Must be squared and
            divisible by 64.
        pretrained: If True, use a pretrained backbone (on COCO).

    # Returns
        A PyTorch model.
    """
    config = get_efficientdet_config(model_name=model_name)

    net = EfficientDet(config, pretrained_backbone=False)
    if pretrained:
        if not config.url:
            raise RuntimeError(f"No pretrained weights for {model_name}")
        state_dict = torch.hub.load_state_dict_from_url(
            config.url, map_location=torch.device("cpu")
        )
        net.load_state_dict(state_dict)

    config.num_classes = num_classes
    config.image_size = img_size
    net.class_net = HeadNet(
        config,
        num_outputs=num_classes,
        norm_kwargs=dict(eps=0.001, momentum=0.01),
    )

    # TODO: Break down param groups for backbone
    def param_groups_fn(model: nn.Module) -> List[List[nn.Parameter]]:
        unwrapped = unwrap_bench(model)

        layers = [
            unwrapped.backbone,
            unwrapped.fpn,
            nn.Sequential(unwrapped.class_net, unwrapped.box_net),
        ]
        param_groups = [list(layer.parameters()) for layer in layers]
        check_all_model_params_in_groups2(model, param_groups)

        return param_groups

    model_bench = DetBenchTrain(net, config)
    model_bench.param_groups = MethodType(param_groups_fn, model_bench)

    return model_bench



__all__ = [
    "train_dl",
    "valid_dl",
    "infer_dl",
    "build_train_batch",
    "build_valid_batch",
    "build_infer_batch",
]

from icevision.imports import *
from icevision.core import *
from icevision.models.utils import *
from icevision.models.rcnn.faster_rcnn.dataloaders import _build_train_sample
from icevision.models.rcnn.faster_rcnn.dataloaders import (
    build_infer_batch,
    infer_dl,
)


def train_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for training the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_train_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def valid_dl(dataset, batch_tfms=None, **dataloader_kwargs) -> DataLoader:
    """A `DataLoader` with a custom `collate_fn` that batches items as required for validating the model.

    # Arguments
        dataset: Possibly a `Dataset` object, but more generally, any `Sequence` that returns records.
        batch_tfms: Transforms to be applied at the batch level.
        **dataloader_kwargs: Keyword arguments that will be internally passed to a Pytorch `DataLoader`.
        The parameter `collate_fn` is already defined internally and cannot be passed here.

    # Returns
        A Pytorch `DataLoader`.
    """
    return transform_dl(
        dataset=dataset,
        build_batch=build_valid_batch,
        batch_tfms=batch_tfms,
        **dataloader_kwargs
    )


def _build_mask_train_sample(record: RecordType):
    assert len(record["labels"]) == len(record["bboxes"]) == len(record["masks"])

    image, target = _build_train_sample(record=record)

    # If no labels and bboxes are present, use as negative samples as described in
    # https://github.com/pytorch/vision/releases/tag/v0.6.0
    if len(record["masks"]) == 0:
        height, width = record["img"].shape[:-1]
        target["masks"] = torch.zeros((0, height, width), dtype=torch.uint8)
    else:
        target["masks"] = tensor(record["masks"].data, dtype=torch.uint8)

    return image, target


def build_train_batch(
    records: List[RecordType], batch_tfms=None
) -> Tuple[List[torch.Tensor], List[Dict[str, torch.Tensor]]]:
    """Builds a batch in the format required by the model when training.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_train_batch(records)
    outs = model(*batch)
    ```
    """
    records = common_build_batch(records)

    images, targets = [], []
    for record in records:
        image, target = _build_mask_train_sample(record)
        images.append(image)
        targets.append(target)

    return (images, targets), records


def build_valid_batch(
    records: List[RecordType], batch_tfms=None
) -> Tuple[List[torch.Tensor], List[Dict[str, torch.Tensor]]]:
    """Builds a batch in the format required by the model when validating.

    # Arguments
        records: A `Sequence` of records.
        batch_tfms: Transforms to be applied at the batch level.

    # Returns
        A tuple with two items. The first will be a tuple like `(images, targets)`,
        in the input format required by the model. The second will be an updated list
        of the input records with `batch_tfms` applied.

    # Examples

    Use the result of this function to feed the model.
    ```python
    batch, records = build_valid_batch(records)
    outs = model(*batch)
    ```
    """
    return build_train_batch(records=records)



__all__ = ["ParserInterface", "Parser"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.parsers.mixins import *


class ParserInterface(ABC):
    @abstractmethod
    def parse(
        self, data_splitter: DataSplitter, show_pbar: bool = True
    ) -> List[List[RecordType]]:
        pass


class Parser(ImageidMixin, ParserInterface, ABC):
    """Base class for all parsers, implements the main parsing logic.

    The actual fields to be parsed are defined by the mixins used when
    defining a custom parser. The only required field for all parsers
    is the `image_id`.

    # Examples

    Create a parser for image filepaths.
    ```python
    class FilepathParser(Parser, FilepathParserMixin):
        # implement required abstract methods
    ```
    """

    def prepare(self, o):
        pass

    @abstractmethod
    def __iter__(self) -> Any:
        pass

    def parse_dicted(
        self, idmap: IDMap, show_pbar: bool = True
    ) -> Dict[int, RecordType]:
        info_parse_funcs = self.collect_info_parse_funcs()
        annotation_parse_funcs = self.collect_annotation_parse_funcs()

        has_invalid_data_error = False

        get_imageid = info_parse_funcs.pop("imageid")

        records = defaultdict(lambda: {name: [] for name in annotation_parse_funcs})
        for sample in pbar(self, show_pbar):
            self.prepare(sample)
            imageid = idmap[get_imageid(sample)]

            for name, func in info_parse_funcs.items():
                records[imageid][name] = func(sample)

            for name, func in annotation_parse_funcs.items():
                try:
                    records[imageid][name].extend(func(sample))
                except InvalidDataError as e:
                    has_invalid_data_error = True
                    true_imageid = idmap.get_id(imageid)
                    logger.info(
                        f"\nInvalid data error in imageid:{true_imageid}\n{str(e)}"
                    )

        if has_invalid_data_error:
            raise InvalidDataError(
                "Some annotations are invalid, check the errors listed here above."
            )

        # check that all annotations have the same length
        # HACK: Masks is not checked, because it can be a single file with multiple masks
        annotations_names = [n for n in annotation_parse_funcs.keys() if n != "masks"]
        for imageid, record_annotations in records.items():
            record_annotations_len = {
                name: len(record_annotations[name]) for name in annotations_names
            }
            if not allequal(list(record_annotations_len.values())):
                true_imageid = idmap.get_id(imageid)
                # TODO: instead of immediatily raising the error, store the
                # result and raise at the end of the for loop for all records
                raise RuntimeError(
                    f"imageid->{true_imageid} has an inconsistent number of annotations"
                    f", all annotations must have the same length."
                    f"\nNumber of annotations: {record_annotations_len}"
                )

        return dict(records)

    def parse(
        self,
        data_splitter: DataSplitter = None,
        idmap: IDMap = None,
        show_pbar: bool = True,
    ) -> List[List[RecordType]]:
        """Loops through all data points parsing the required fields.

        # Arguments
            data_splitter: How to split the parsed data, defaults to a [0.8, 0.2] random split.
            idmap: Maps from filenames to unique ids, pass an `IDMap()` if you need this information.
            show_pbar: Whether or not to show a progress bar while parsing the data.

        # Returns
            A list of records for each split defined by `data_splitter`.
        """
        idmap = idmap or IDMap()
        data_splitter = data_splitter or RandomSplitter([0.8, 0.2])
        records = self.parse_dicted(show_pbar=show_pbar, idmap=idmap)
        splits = data_splitter(idmap=idmap)
        return [[{"imageid": id, **records[id]} for id in ids] for ids in splits]

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return ["def __iter__(self) -> Any:"] + templates

    @classmethod
    def generate_template(cls):
        for template in cls._templates():
            print(f"{template}")



__all__ = ["Metric"]

from icevision.imports import *


class Metric(ABC):
    def __init__(self):
        self._model = None

    @abstractmethod
    def accumulate(self, records, preds):
        """Accumulate stats for a single batch"""

    @abstractmethod
    def finalize(self) -> Dict[str, float]:
        """Called at the end of the validation loop"""

    @property
    def name(self) -> str:
        return self.__class__.__name__



__all__ = [
    "ImageidMixin",
    "FilepathMixin",
    "SizeMixin",
    "LabelsMixin",
    "BBoxesMixin",
    "AreasMixin",
    "MasksMixin",
    "IsCrowdsMixin",
]

from icevision.imports import *
from icevision.core import *


class ParserMixin(ABC):
    def collect_annotation_parse_funcs(self, funcs=None):
        return funcs or {}

    def collect_info_parse_funcs(self, funcs=None):
        return funcs or {}

    @classmethod
    def generate_template(cls):
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        return []


class ImageidMixin(ParserMixin):
    """Adds `imageid` method to parser"""

    def collect_info_parse_funcs(self, funcs=None):
        funcs = super().collect_info_parse_funcs(funcs)
        return {"imageid": self.imageid, **funcs}

    @abstractmethod
    def imageid(self, o) -> Hashable:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def imageid(self, o) -> Hashable:"]


class FilepathMixin(ParserMixin):
    """Adds `filepath` method to parser"""

    def collect_info_parse_funcs(self, funcs=None):
        funcs = super().collect_info_parse_funcs(funcs)
        return {"filepath": self.filepath, **funcs}

    @abstractmethod
    def filepath(self, o) -> Union[str, Path]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def filepath(self, o) -> Union[str, Path]:"]


class SizeMixin(ParserMixin):
    """Adds `image_height` and `image_width` method to parser"""

    def collect_info_parse_funcs(self, funcs=None):
        funcs = super().collect_info_parse_funcs(funcs)
        return {"height": self.image_height, "width": self.image_width, **funcs}

    @abstractmethod
    def image_height(self, o) -> int:
        pass

    @abstractmethod
    def image_width(self, o) -> int:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + [
            "def image_height(self, o) -> int:",
            "def image_width(self, o) -> int:",
        ]


### Annotation parsers ###
class LabelsMixin(ParserMixin):
    """Adds `labels` method to parser"""

    def collect_annotation_parse_funcs(self, funcs=None):
        funcs = super().collect_annotation_parse_funcs(funcs)
        return {"labels": self.labels, **funcs}

    @abstractmethod
    def labels(self, o) -> List[int]:
        """Returns the labels for the receive sample.

        !!! danger "Important"
        If you are using a RCNN/efficientdet model,
        remember to return 0 for background.
        """

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def labels(self, o) -> List[int]:"]


class BBoxesMixin(ParserMixin):
    """Adds `bboxes` method to parser"""

    def collect_annotation_parse_funcs(self, funcs=None):
        funcs = super().collect_annotation_parse_funcs(funcs)
        return {"bboxes": self.bboxes, **funcs}

    @abstractmethod
    def bboxes(self, o) -> List[BBox]:
        pass

    @classmethod
    def generate_template(cls):
        print("def bboxes(self, o) -> List[BBox]:")
        super().generate_template()

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def bboxes(self, o) -> List[BBox]:"]


class MasksMixin(ParserMixin):
    """Adds `masks` method to parser"""

    def collect_annotation_parse_funcs(self, funcs=None):
        funcs = super().collect_annotation_parse_funcs(funcs)
        return {"masks": self.masks, **funcs}

    @abstractmethod
    def masks(self, o) -> List[Mask]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def masks(self, o) -> List[Mask]:"]


class AreasMixin(ParserMixin):
    """Adds `areas` method to parser"""

    def collect_annotation_parse_funcs(self, funcs=None):
        funcs = super().collect_annotation_parse_funcs(funcs)
        return {"areas": self.areas, **funcs}

    @abstractmethod
    def areas(self, o) -> List[float]:
        """Returns areas of the segmentation"""

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def areas(self, o) -> List[float]:"]


class IsCrowdsMixin(ParserMixin):
    """Adds `iscrowds` method to parser"""

    def collect_annotation_parse_funcs(self, funcs=None):
        funcs = super().collect_annotation_parse_funcs(funcs)
        return {"iscrowds": self.iscrowds, **funcs}

    @abstractmethod
    def iscrowds(self, o) -> List[bool]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def iscrowds(self, o) -> List[bool]:"]



# The draw functions are based from:
# https://github.com/fizyr/keras-retinanet/blob/master/keras_retinanet/utils/visualization.py
# https://github.com/fizyr/keras-maskrcnn/blob/master/keras_maskrcnn/utils/visualization.py

__all__ = [
    "draw_sample",
    "draw_record",
    "draw_pred",
    "draw_bbox",
    "draw_mask",
]

from icevision.imports import *
from icevision.data import *
from icevision.core import *


def draw_sample(
    sample,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
):
    img = sample["img"].copy()
    if denormalize_fn is not None:
        img = denormalize_fn(img)

    for label, bbox, mask in itertools.zip_longest(
        sample.get("labels", []), sample.get("bboxes", []), sample.get("masks", [])
    ):
        color = (np.random.random(3) * 0.6 + 0.4) * 255

        if display_mask and mask is not None:
            img = draw_mask(img=img, mask=mask, color=color)
        if display_bbox and bbox is not None:
            img = draw_bbox(img=img, bbox=bbox, color=color)
        if display_label and label is not None:
            img = draw_label(
                img=img,
                label=label,
                bbox=bbox,
                mask=mask,
                class_map=class_map,
                color=color,
            )

    return img


def draw_label(
    img: np.ndarray,
    label: int,
    color,
    class_map: Optional[ClassMap] = None,
    bbox=None,
    mask=None,
):
    # finds label position based on bbox or mask
    if bbox is not None:
        x, y = bbox.x, bbox.y
    elif mask is not None:
        y, x = np.unravel_index(mask.data.argmax(), mask.data.shape)
    else:
        x, y = 0, 0

    if class_map is not None:
        caption = class_map.get_id(label)
    else:
        caption = str(label)

    return _draw_label(img=img, caption=caption, x=int(x), y=int(y), color=color)


def _draw_label(
    img: np.ndarray,
    caption: str,
    x: int,
    y: int,
    color,
    font=cv2.FONT_HERSHEY_SIMPLEX,
    font_scale: float = 1.0,
):
    """Draws a caption above the box in an image."""
    y -= 10
    w, h = cv2.getTextSize(caption, font, fontScale=font_scale, thickness=1)[0]

    # make the coords of the box with a small padding of two pixels
    box_pt1, box_pt2 = (x, y + 10), (x + w + 2, y - h - 2)
    cv2.rectangle(img, box_pt1, box_pt2, color, cv2.FILLED)

    cv2.putText(img, caption, (x, y), font, font_scale, (240, 240, 240), 2)

    return img


def draw_record(
    record,
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    prepare_record: Optional[callable] = None,
):
    data_preparer = prepare_record or default_prepare_record
    sample = data_preparer(record)
    return draw_sample(
        sample=sample,
        class_map=class_map,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
    )


def draw_pred(
    img: np.ndarray,
    pred: dict,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
):
    sample = pred.copy()
    sample["img"] = img
    return draw_sample(
        sample=sample,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
    )


def draw_bbox(
    img: np.ndarray, bbox: BBox, color: Tuple[int, int, int], thickness: int = 2
):
    """Draws a box on an image with a given color.
    # Arguments
        image     : The image to draw on.
        box       : A list of 4 elements (x1, y1, x2, y2).
        color     : The color of the box.
        thickness : The thickness of the lines to draw a box with.
    """
    xyxy = tuple(np.array(bbox.xyxy, dtype=int))
    cv2.rectangle(img, xyxy[:2], xyxy[2:], color, thickness, cv2.LINE_AA)
    return img


def draw_mask(
    img: np.ndarray, mask: MaskArray, color: Tuple[int, int, int], blend: float = 0.5
):
    color = np.asarray(color, dtype=int)
    # draw mask
    mask_idxs = np.where(mask.data)
    img[mask_idxs] = blend * img[mask_idxs] + (1 - blend) * color

    # draw border
    border = mask.data - cv2.erode(mask.data, np.ones((7, 7), np.uint8), iterations=1)
    border_idxs = np.where(border)
    img[border_idxs] = color

    return img



__all__ = ["show_results"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.base_show_results import base_show_results
from icevision.models.rcnn.faster_rcnn.dataloaders import build_infer_batch
from icevision.models.rcnn.faster_rcnn.prediction import predict


def show_results(
    model: nn.Module,
    dataset: Dataset,
    class_map: Optional[ClassMap] = None,
    num_samples: int = 6,
    ncols: int = 3,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    show: bool = True,
) -> None:
    return base_show_results(
        predict_fn=predict,
        build_infer_batch_fn=build_infer_batch,
        model=model,
        dataset=dataset,
        class_map=class_map,
        num_samples=num_samples,
        ncols=ncols,
        denormalize_fn=denormalize_fn,
        show=show,
    )



__all__ = ["RCNNCallback", "rcnn_learner"]

from icevision.imports import *
from icevision.engines.fastai import *
from icevision.models.rcnn.loss_fn import loss_fn
from icevision.models.rcnn.fastai.callbacks import *


def rcnn_learner(
    dls: List[Union[DataLoader, fastai.DataLoader]],
    model: nn.Module,
    cbs=None,
    **kwargs
):
    learn = adapted_fastai_learner(
        dls=dls,
        model=model,
        cbs=cbs,
        loss_func=loss_fn,
        **kwargs,
    )

    # HACK: patch AvgLoss (in original, find_bs gives errors)
    class RCNNAvgLoss(fastai.AvgLoss):
        def accumulate(self, learn):
            bs = len(learn.yb)
            self.total += fastai.to_detach(learn.loss.mean()) * bs
            self.count += bs

    recorder = [cb for cb in learn.cbs if isinstance(cb, fastai.Recorder)][0]
    recorder.loss = RCNNAvgLoss()

    return learn



__all__ = ["adapted_fastai_learner"]

from icevision.imports import *
from icevision.utils import *
from icevision.metrics import *
from icevision.engines.fastai.imports import *
from icevision.engines.fastai.adapters import *


# TODO: param_groups fix for efficientdet
def adapted_fastai_learner(
    dls: List[Union[DataLoader, fastai.DataLoader]],
    model: nn.Module,
    metrics=None,
    device=None,
    splitter=None,
    **learner_kwargs,
) -> fastai.Learner:
    # convert dataloaders to fastai
    fastai_dls = []
    for dl in dls:
        if isinstance(dl, DataLoader):
            fastai_dl = convert_dataloader_to_fastai(dl)
        elif isinstance(dl, fastai.DataLoader):
            fastai_dl = dl
        else:
            raise ValueError(f"dl type {type(dl)} not supported")

        fastai_dls.append(fastai_dl)

    device = device or fastai.default_device()
    fastai_dls = fastai.DataLoaders(*fastai_dls).to(device)

    # convert metrics to fastai
    metrics = metrics or []
    fastai_metrics = [
        FastaiMetricAdapter(metric) if isinstance(metric, Metric) else metric
        for metric in metrics
    ]

    if splitter == None:
        if hasattr(model, "param_groups"):

            def splitter(model):
                return model.param_groups()

        else:
            raise ValueError(
                "If the parameter `splitter` is not specified, "
                "the model should define a method called `param_groups`"
            )

    return fastai.Learner(
        dls=fastai_dls,
        model=model,
        metrics=fastai_metrics,
        splitter=splitter,
        **learner_kwargs,
    )



__all__ = ["BatchTransform"]

from icevision.imports import *
from icevision.core import *


class BatchTransform(ABC):
    def __call__(self, records: List[RecordType]) -> List[RecordType]:
        return self.apply(records=records)

    @abstractmethod
    def apply(self, records: List[RecordType]) -> List[RecordType]:
        """Main transform logic.

        Args:
            records (List[RecordType]): List of records to apply transforms.
        """



__all__ = [
    "download_url",
    "download_and_extract",
    "download_gdrive",
    "download_and_extract_gdrive",
]

from icevision.imports import *
import requests


def download_url(url, save_path, chunk_size=1024) -> None:
    """Download file from url"""
    r = requests.get(url, stream=True)
    with open(save_path, "wb") as f:
        bar_total = r.headers.get("Content-Length")
        bar = tqdm(unit="B", total=int(bar_total) if bar_total else None)
        for chunk in r.iter_content(chunk_size=chunk_size):
            if chunk:  # filter out keep-alive new chunks
                bar.update(len(chunk))
                f.write(chunk)


def download_and_extract(
    url: Union[str, Path], save_path: Union[str, Path], chunk_size: int = 1024
) -> None:
    save_dir = Path(save_path).parent

    download_url(url=str(url), save_path=str(save_path), chunk_size=chunk_size)
    shutil.unpack_archive(filename=str(save_path), extract_dir=str(save_dir))


def download_gdrive(url, save_path):
    """Download from gdrive, passing virus scan for big files."""
    import gdown

    gdown.download(url=str(url), output=str(save_path), quiet=False)


def download_and_extract_gdrive(url, filename, extract_dir):
    extract_dir = Path(extract_dir)
    save_path = extract_dir / filename
    extract_dir.mkdir(exist_ok=True, parents=True)

    download_gdrive(url=url, save_path=save_path)
    shutil.unpack_archive(filename=str(save_path), extract_dir=str(extract_dir))



__all__ = [
    "to_np",
    "requires_grad",
    "model_device",
    "params",
    "check_all_model_params_in_groups2",
    "model_device",
]

from icevision.imports import *


def to_np(t):
    return t.detach().cpu().numpy()


def requires_grad(model, layer):
    return list(model.parameters())[layer].requires_grad


def model_device(model):
    return first(model.parameters()).device


def params(m):
    return list(m.parameters())


def check_all_model_params_in_groups2(
    model: nn.Module, param_groups: List[List[nn.Parameter]]
):
    num_params = len([param for group in param_groups for param in group])
    num_params_expected = len(list(model.parameters()))

    if num_params != num_params_expected:
        raise RuntimeError(
            f"{num_params_expected} params in model but only {num_params} "
            "in parameter group"
        )


def model_device(model: nn.Module):
    """Returns the device the first model parameter is stored.

    Can be wrong if different parts of the model are in different devices.
    """
    return next(iter(model.parameters())).device



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['icevision',
 'icevision.backbones',
 'icevision.core',
 'icevision.data',
 'icevision.engines',
 'icevision.engines.fastai',
 'icevision.engines.fastai.adapters',
 'icevision.engines.fastai.learner',
 'icevision.engines.lightning',
 'icevision.metrics',
 'icevision.metrics.coco_metric',
 'icevision.models',
 'icevision.models.efficientdet',
 'icevision.models.efficientdet.fastai',
 'icevision.models.efficientdet.lightning',
 'icevision.models.rcnn',
 'icevision.models.rcnn.fastai',
 'icevision.models.rcnn.faster_rcnn',
 'icevision.models.rcnn.faster_rcnn.fastai',
 'icevision.models.rcnn.faster_rcnn.lightning',
 'icevision.models.rcnn.lightning',
 'icevision.models.rcnn.mask_rcnn',
 'icevision.models.rcnn.mask_rcnn.fastai',
 'icevision.models.rcnn.mask_rcnn.lightning',
 'icevision.parsers',
 'icevision.parsers.mixins',
 'icevision.test_utils',
 'icevision.tfms',
 'icevision.tfms.albumentations',
 'icevision.tfms.batch',
 'icevision.utils',
 'icevision.visualize']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=0.4.6,<0.5.0',
 'fastcore>=1.0.14,<2.0.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.3.2,<4.0.0',
 'opencv-python>=4.4.0,<5.0.0',
 'pycocotools>=2.0.2,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'torch>=1.6.0,<1.7.0',
 'torchvision>=0.7.0,<0.8.0',
 'tqdm>=4.49.0,<5.0.0']

extras_require = \
{'all': ['fastai>=2.0.13',
         'ipykernel>=5.3.4,<6.0.0',
         'pytorch-lightning>=0.9,<0.10',
         'effdet>=0.1.6,<0.2.0',
         'omegaconf>=2.0.0,<3.0.0',
         'dataclasses==0.6'],
 'fastai': ['fastai>=2.0.13', 'ipykernel>=5.3.4,<6.0.0'],
 'inference': ['effdet>=0.1.6,<0.2.0',
               'omegaconf>=2.0.0,<3.0.0',
               'dataclasses==0.6'],
 'lightning': ['pytorch-lightning>=0.9,<0.10'],
 'models': ['effdet>=0.1.6,<0.2.0',
            'omegaconf>=2.0.0,<3.0.0',
            'dataclasses==0.6'],
 'training': ['fastai>=2.0.13',
              'ipykernel>=5.3.4,<6.0.0',
              'pytorch-lightning>=0.9,<0.10',
              'effdet>=0.1.6,<0.2.0',
              'omegaconf>=2.0.0,<3.0.0',
              'dataclasses==0.6']}

setup_kwargs = {
    'name': 'icevision',
    'version': '0.2.0b1',
    'description': 'Agnostic Computer Vision Framework',
    'long_description': None,
    'author': 'Lucas Goulart Vazquez',
    'author_email': 'lgvaz@airctic.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from icevision.models.efficientdet.model import *
from icevision.models.efficientdet.dataloaders import *
from icevision.models.efficientdet.loss_fn import *
from icevision.models.efficientdet.prediction import *
from icevision.models.efficientdet.show_results import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import icevision.models.efficientdet.fastai

if SoftDependencies.pytorch_lightning:
    import icevision.models.efficientdet.lightning



__all__ = ["vgg_param_groups", "vgg11", "vgg13", "vgg16", "vgg19"]

from icevision.imports import *
from icevision.utils import *


def _vgg_features(model: nn.Module):
    features = model.features
    features.out_channels = 512
    features.param_groups = MethodType(vgg_param_groups, features)

    return features


def vgg_param_groups(model: nn.Module) -> List[List[nn.Parameter]]:
    layers = []
    layers += [model[:4]]
    # splits layers into 3 equally sized chunks
    for group in np.array_split(model[4:], 3):
        layers += [nn.Sequential(*group)]

    param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, param_groups)

    return param_groups


def vgg11(pretrained: bool = True):
    return _vgg_features(model=torchvision.models.vgg11(pretrained=pretrained))


def vgg13(pretrained: bool = True):
    return _vgg_features(model=torchvision.models.vgg13(pretrained=pretrained))


def vgg16(pretrained: bool = True):
    return _vgg_features(model=torchvision.models.vgg16(pretrained=pretrained))


def vgg19(pretrained: bool = True):
    return _vgg_features(model=torchvision.models.vgg19(pretrained=pretrained))



__all__ = ["ModelAdapter"]

from icevision.models.rcnn.lightning.model_adapter import *
from icevision.models.rcnn.faster_rcnn.prediction import *


class ModelAdapter(RCNNModelAdapter):
    """Lightning module specialized for faster_rcnn, with metrics support.

    The methods `forward`, `training_step`, `validation_step`, `validation_epoch_end`
    are already overriden.

    # Arguments
        model: The pytorch model to use.
        metrics: `Sequence` of metrics to use.

    # Returns
        A `LightningModule`.
    """

    def convert_raw_predictions(self, raw_preds):
        return convert_raw_predictions(raw_preds=raw_preds, detection_threshold=0.0)



__all__ = ["get_files", "get_image_files"]

from icevision.imports import *

# All copied from fastai
def _get_files(p, fs, extensions=None):
    p = Path(p)
    res = [
        p / f
        for f in fs
        if not f.startswith(".")
        and ((not extensions) or f'.{f.split(".")[-1].lower()}' in extensions)
    ]
    return res


def get_files(
    path,
    extensions=None,
    recurse=True,
    folders=None,
    followlinks=True,
    sort: bool = True,
):
    "Get all the files in `path` with optional `extensions`, optionally with `recurse`, only in `folders`, if specified. From fastai"
    path = Path(path)
    folders = L(folders)
    extensions = setify(extensions)
    extensions = {e.lower() for e in extensions}
    if recurse:
        res = []
        for i, (p, d, f) in enumerate(
            os.walk(path, followlinks=followlinks)
        ):  # returns (dirpath, dirnames, filenames)
            if len(folders) != 0 and i == 0:
                d[:] = [o for o in d if o in folders]
            else:
                d[:] = [o for o in d if not o.startswith(".")]
            if len(folders) != 0 and i == 0 and "." not in folders:
                continue
            res += _get_files(p, f, extensions)
    else:
        f = [o.name for o in os.scandir(path) if o.is_file()]
        res = _get_files(path, f, extensions)

    return L(sorted(res)) if sort else L(res)


image_extensions = set(
    k for k, v in mimetypes.types_map.items() if v.startswith("image/")
)


def get_image_files(path, recurse=True, folders=None):
    "Get image files in `path` recursively, only in `folders`, if specified. From fastai"
    return get_files(
        path, extensions=image_extensions, recurse=recurse, folders=folders
    )



__all__ = ["RCNNModelAdapter"]

from icevision.imports import *
from icevision.utils import *
from icevision.metrics import *
from icevision.engines.lightning.lightning_model_adapter import LightningModelAdapter
from icevision.models.rcnn.loss_fn import loss_fn


class RCNNModelAdapter(LightningModelAdapter, ABC):
    def __init__(self, model: nn.Module, metrics: Sequence[Metric] = None):
        super().__init__(metrics=metrics)
        self.model = model

    @abstractmethod
    def convert_raw_predictions(self, raw_preds):
        """Convert raw predictions from the model to library standard."""

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def training_step(self, batch, batch_idx):
        (xb, yb), records = batch
        preds = self(xb, yb)

        loss = loss_fn(preds, yb)
        log = {"train/loss": loss}

        return {"loss": loss, "log": log}

    def validation_step(self, batch, batch_idx):
        (xb, yb), records = batch

        with torch.no_grad():
            self.train()
            train_preds = self(xb, yb)
            loss = loss_fn(train_preds, yb)

            self.eval()
            raw_preds = self(xb)
            preds = self.convert_raw_predictions(raw_preds=raw_preds)
            self.accumulate_metrics(records=records, preds=preds)

        return {"valid/loss": loss}

    def validation_epoch_end(self, outs):
        loss_log = {k: torch.stack(v).mean() for k, v in mergeds(outs).items()}
        metrics_log = self.finalize_metrics()
        log = {**loss_log, **metrics_log}
        return {"val_loss": log["valid/loss"], "log": log}



from icevision.core.exceptions import *
from icevision.core.id_map import *
from icevision.core.class_map import *
from icevision.core.mask import *
from icevision.core.bbox import *
from icevision.core.record_type import *
